# -*- coding: utf-8 -*-

#########################################################################
# pyeole.service - manage EOLE services
# Copyright © 2012-2013 Pôle de Compétence EOLE <eole@ac-dijon.fr>
#
# License CeCILL:
#  * in french: http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
#  * in english http://www.cecill.info/licences/Licence_CeCILL_V2-en.html
#########################################################################

"""EOLE start/stop/restart/reload service or install/remove init script

Start, stop or restart available service
----------------------------------------

Generic interface to start, stop, restart or reload services in Creole
environment.

This implementation support two types:

* system V style init script
* upstart

All other service methods are ignored.

Examples:

    >>> from pyeole.log import init_logging
    >>> log = init_logging(level='info')
    >>> from pyeole.service import manage_services
    >>> try:
    ...     manage_services('start', 'servicename')
    ... except UnknownServiceError, err:
    ...     log.error(err)

Add service in boot sequence
----------------------------

There are several init script types. This implementation support
different's types:

* system V style init script
* upstart
* special services (like apache, ...)

Examples:

    >>> from pyeole.service import manage_services
    >>> manage_services('enable', 'servicename')
    >>> manage_services('disable', 'unwantedservice')

"""

import re

from os import symlink

from os.path import dirname
from os.path import join
from os.path import isfile
from os.path import relpath

from glob import glob

from importlib import import_module

from creole.client import CreoleClient
from creole.client import NotFoundError
from creole.client import CreoleClientError

from pyeole.decorator import deprecated

# Base exception
from pyeole.service.error import ServiceError
# Configuration
from pyeole.service.error import DisabledError
from pyeole.service.error import ConfigureError
from pyeole.service.error import UnknownServiceError

from pyeole.service.launcher import get_commands

import logging

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

_ENCODING = 'utf-8'
"""Force encoding to UTF-8

"""

_CLIENT = CreoleClient()


def manage_services(action, names=None, container=None, display="log", exclude=None, force=False):
    """Apply :data:`action` to one or several services.

    If no service :data:`names` is provided, :data:`action` is
    performed for all services.

    if no :data:`container` is provided, services are looked-up in all
    container groups.

    :param action: action to perform
    :type action: `str` in [``configure``, ``apply``, ``enable``,
                  ``disable``, ``status``, ``start``, ``restart``,
                  ``stop``, ``reload``]
    :param names: names of services
    :type names: `list` of `str`
    :param container: name of the container where to perform :data:`action`
    :type container: `str`
    :param display: type of message display "log", "console", "both"
    :type display: `str`
    :param exclude: exclude services
    :type exclude: `list`
    :param force: don't validate Creole state of service
    :type force: `bool`

    """
    if not force and action in ['enable', 'disable']:
        raise ConfigureError('cannot enable, disable directly use configure instead')

    containers_ctx = []
    if container is not None:
        containers_ctx = [_CLIENT.get_container_infos(container)]
    else:
        for group_name in _CLIENT.get_groups():
            containers_ctx.append(_CLIENT.get_group_infos(group_name))

    if names is not None and not isinstance(names, list):
        names = [names]

    # Check if service is unknown or disabled in all containers
    if names is None:
        services_str = u'all services'
    else:
        services_str = u', '.join(names)

        for name in names:
            service_list = _CLIENT.get_service(name, container)
            if service_list == []:
                msg = u'Service unknown to creoled: {0}'.format(name)
                raise UnknownServiceError(msg)

            if not force and action in ['start', 'restart', 'reload']:
                enabled = False
                # Disabled in all containers?
                for srv in service_list:
                    enabled |= srv.get('activate', False)
                if not enabled:
                    msg = 'cannot start disabled service'
                    raise DisabledError(msg)

    log.debug(u'Run {0} for {1}'.format(action, services_str))
    commands = get_commands(display)
    err_found = False

    for container in containers_ctx:
        services = []

        if names is None:
            services = container['services']
        else:
            # This code keep order of creole dictionaries
            for service in container[u'services']:
                if service[u'name'] in names:
                    services.append(service)

        if action == u'stop':
            services.reverse()

        for service in services:

            if action in ['start', 'restart', 'reload'] and \
               not service.get('activate', False) and not force:
                msg = 'cannot start disabled service'
                log.debug(msg)
                continue

            if exclude is None or (
                    (service[u'container'], service[u'name']) not in exclude and \
                    (container[u'name'] , service[u'name']) not in exclude):
                ret = _manage_service(action,
                                      service=service,
                                      ctx=container,
                                      command=commands[service['method']])
                # on regarde si _manage_service a renvoyé un code de retour
                if ret and ret.get('code', 0) != 0:
                    err_found = True

    return int(err_found)


def unmanaged_service(action, name, method, container='root', display='log', force=False):
    """Apply :data:`action` to a service not managed in creole.

    Prepare the service and context dictionaries to call
    :func:`_manage_service`.

    :param action: action to perform
    :type action: `str` in [``start``, ``restart``, ``stop``]
    :param name: name of service
    :type name: `str`
    :param method: method to manage service in [``service``, ``upstart``]
    :type metho: `str`
    :param container: name of the container where to perform :data:`action`
    :type container: `str`
    :param display: type of message display "log", "console", "both"
    :type display: `str`
    :param force: allow disable and enable action
    :type force: `boolean`

    """
    if method not in [u'service', u'upstart', u'network', u'apache']:
        msg = u'Unauthorized service method for unmanaged service: {0}'
        raise ServiceError(msg.format(method))

    if action not in [u'start', u'restart', u'stop']:
        if force and action in [u'disable', u'enable']:
            pass
        else:
            msg = u'Unauthorized action for unmanaged service: {0}'
            raise ServiceError(msg.format(action))

    service = {u'name': name,
               u'method': method,
               u'activate': True,
               u'container': container,
               u'pty': True}

    if container == 'root' and name == 'creoled':
        ctx = {u'name': container,
               u'ip': u'127.0.0.1',
               u'path': '/'}
    else:
        try:
            ctx = _CLIENT.get_container(container)
        except CreoleClientError, err:
            # Avoid error if creoled is not running
            # but only for root container
            raise err


    commands = get_commands(display)
    # unmanaged_services: raises an exception if underlying code returns {'code':0, 'msg':err_msg}
    ret = _manage_service(action, service, ctx, command=commands[method])
    if ret and ret.get('code', 0) != 0:
        raise ServiceError(ret['msg'])


#####
##### Global worker
#####


def _manage_service(action, service, ctx, command):
    """Apply :data:`action` to :data:`services` in a container.

    :param action: action to perform
    :type action: `str` in [``configure``, ``apply``, ``enable``,
                  ``disable``, ``status``, ``start``, ``restart``,
                  ``stop``, ``reload``]
    :param services: names of service
    :type services: `list` of `str`
    :param ctx: container context
    :type ctx: `dict`
    :raise ServiceError: if the action is not valid
    :raise UnknownServiceError: if the service is unknown to creoled

    """
    if action in ['start', 'restart', 'reload', 'configure']:
        command.check_service(service, ctx)
    return getattr(command, action + '_service')(service, ctx)


##########
########## OLD COLD FROM HERE
##########

##########
########## Deprecated API
##########

@deprecated(u'Use new API “manage_services()”')
def service_all(action, network=False, lxc=True):
    """Old API to run :data:`action` on all services

    Use the new :func:`manage_services` API.

    """
    manage_services(action)


@deprecated(u'Use new API “manage_services()”')
def service_code(service, action, container=u'root'):
    """Old API to run :data:`action` on a service in a container

    Use the new :func:`manage_services` API.

    """
    manage_services(action, service, container=container)


@deprecated(u'Use new API “manage_services()”')
def creole_service_code(service, action, container):
    """Old API to run :data:`action` on a service in a container

    Use the new :func:`manage_services` API.

    """
    manage_services(action, service[u'name'], container=container[u'name'])


def service_out(service, action, container=u'root'):
    """Old broken API to run :data:`action` on a service in a container

    Use the new :func:`manage_services` API.

    """
    msg = u'This code was broken, use “manage_services()” API'
    raise NotImplementedError(msg)


@deprecated(u'Use new API “manage_services()”')
def creole_service_out(service, action, container):
    """Old broken API to run :data:`action` on a service in a container

    Use the new :func:`manage_services` API.

    """
    manage_services(action, service[u'name'], container=container[u'name'])
    # Fake return value, manage_services raise if something goes wrong
    return 0, u'', u''


@deprecated(u'Use new API “manage_services()”')
def update_rcd(service, action, container=u'root'):
    """Old API to configure service in boot sequence.

    Use the new :func:`manage_services` API.

    """
    manage_services(action, service, container=container)


@deprecated(u'Use new API “manage_services()”')
def creole_update_rcd(service, action, container):
    """Old API to configure service in boot sequence.

    Use the new :func:`manage_services` API.

    """
    manage_services(action, service[u'name'], container=container[u'name'])


@deprecated(u'Use new API “manage_services()”')
def update_rcd_all():
    """Old API to configure services in boot sequence.

    Use the new :func:`manage_services` API.

    """
    manage_services(u'configure')


@deprecated(u'Use new API "manage_services()"')
def manage_service(action, name=None, container=None, display='log'):
    """Old API to start services.

    Use the new :func:`manage_services` API.

    """
    manage_services(action, names=name, container=container, display=display)
